--[[
Name: InventoryLib-2.0
Revision: $Rev: 46135 $
Author(s): Elkano (elkano@gmx.de)
Website: http://elkano.wowinterface.com/
Documentation: http://wiki.wowace.com/index.php/InventoryLib-2.0
SVN: http://svn.wowace.com/root/trunk/InventoryLib/InventoryLib-2.0
Description: A library that maintains a cache of linkable items and shares it between clients.
Dependencies: AceLibrary, AceComm-2.0, AceEvent-2.0, Gratuity-2.0
]]

local MAJOR_VERSION = "InventoryLib-2.0"
local MINOR_VERSION = "$Revision: 46135 $"

local _G = getfenv(0)

local AceLibrary = _G.AceLibrary
if not AceLibrary then error(MAJOR_VERSION .. " requires AceLibrary") end
if not AceLibrary:IsNewVersion(MAJOR_VERSION, MINOR_VERSION) then return end
if not AceLibrary:HasInstance("AceEvent-2.0") then error(MAJOR_VERSION .. " requires AceEvent-2.0") end

local AceEvent

local PATTERN_itemid = "item:(%-?%d+):%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+"
local PATTERN_itemlink = "|c%x+|H"..PATTERN_itemid.."|h%[.-%]|h|r"
local FORMAT_itemid = "item:%d:0:0:0:0:0:0:0"

local ipairs				= _G.ipairs
local mod					= _G.mod
local pairs					= _G.pairs
local select				= _G.select
local time					= _G.time
local tonumber				= _G.tonumber
local type					= _G.type

local CheckInteractDistance			= _G.CheckInteractDistance
local GetAuctionItemLink			= _G.GetAuctionItemLink
local GetContainerItemLink			= _G.GetContainerItemLink
local GetContainerNumSlots			= _G.GetContainerNumSlots
local GetCraftInfo					= _G.GetCraftInfo
local GetCraftItemLink				= _G.GetCraftItemLink
local GetCraftNumReagents			= _G.GetCraftNumReagents
local GetCraftReagentInfo			= _G.GetCraftReagentInfo
local GetCraftReagentItemLink		= _G.GetCraftReagentItemLink
local GetInventoryItemLink			= _G.GetInventoryItemLink
local GetInventorySlotInfo			= _G.GetInventorySlotInfo
local GetItemInfo					= _G.GetItemInfo
local GetMerchantItemLink			= _G.GetMerchantItemLink
local GetMerchantNumItems			= _G.GetMerchantNumItems
local GetNumAuctionItems			= _G.GetNumAuctionItems
local GetNumCrafts					= _G.GetNumCrafts
local GetNumTradeSkills				= _G.GetNumTradeSkills
local GetTradeSkillInfo				= _G.GetTradeSkillInfo
local GetTradeSkillItemLink			= _G.GetTradeSkillItemLink
local GetTradeSkillNumReagents		= _G.GetTradeSkillNumReagents
local GetTradeSkillReagentInfo		= _G.GetTradeSkillReagentInfo
local GetTradeSkillReagentItemLink	= _G.GetTradeSkillReagentItemLink
local UnitExists					= _G.UnitExists
local UnitIsFriend					= _G.UnitIsFriend
local UnitIsPlayer					= _G.UnitIsPlayer
local UnitName						= _G.UnitName

local band					= _G.bit.band

local math_floor			= _G.math.floor
local math_random			= _G.math.random

local strbyte				= _G.string.byte
local strchar				= _G.string.char
local strfind				= _G.string.find
local strlen				= _G.string.len

local table_concat			= _G.table.concat
local table_insert			= _G.table.insert
local table_remove			= _G.table.remove
local table_sort			= _G.table.sort

local Gratuity
local InventoryLib = {}
-- -----
-- the following functions provide some tools
-- -----
local function getbaseid(link)
	if link then
		if type(link) == "number" then return link end
		local _, _, baseid = link:find(PATTERN_itemlink)
		if not baseid then
			_, _, baseid = link:find(PATTERN_itemid)
		end
		if not baseid then
			baseid = link
		end
		return tonumber(baseid)
	end
	return
end

-- -----
-- public functions
-- -----

-- returns a table of all itemIDs in cache
-- the list is rescanned step by step in background or if forcedRescan is set
function InventoryLib:GetItems(forceRescan)
	if forceRescan then
		self:BuildLists(true)
	end
	return self.localvars.items
end

-- returns true if the item from link is in the client's cache
function InventoryLib:IsCacheItem(item)
	item = getbaseid(item)
	self:BuildLists(false, item)
	return (self.localvars.items[item] ~= nil)
end

-- returns true if the item from link is known as valid for this session
function InventoryLib:IsSessionItem(item)
	item = getbaseid(item)
	self:BuildLists(false, item)
	return (self.localvars.items[item] == true)
end

function InventoryLib:GetTypesTables()
	local localvars = self.localvars
	return localvars.types_tree, localvars.types_base, localvars.types_sub, localvars.types_slot
end

function InventoryLib:GetCounts()
	local localvars = self.localvars
	return localvars.count_known, localvars.count_safe
end

function InventoryLib:GetKnownUsers()
	local localvars = self.localvars
	return localvars.known_users, localvars.known_users_rev
end
-- -----
-- 'internal' functions
-- -----

-- this function scans the client's cache for items
-- it will also build a tree for the various types an item can have and sortet lists for the types and subtypes
function InventoryLib:BuildLists(forceRescan, scanid)
	self:RecheckSessionItems()

	local localvars = self.localvars

	local newtype = false
	local first, last
	if forceRescan == true then
		first = 1
		last = self.consts.maxid
		localvars.items_scan_last = 0
		scanid = nil
	else
		if scanid then
			first = scanid
			last = scanid
		else
			first = localvars.items_scan_last + 1
			last = localvars.items_scan_last + self.consts.itemscancount
			localvars.items_scan_last = (last < self.consts.maxid) and last or 0
		end
	end
	local types_tree = localvars.types_tree
	local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc
	for i = first, last do
		if localvars.items[i] == nil or scanid then
			itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc = GetItemInfo(i)
			if itemName then	-- valid item
				itemEquipLoc = _G[itemEquipLoc]
				if (not itemEquipLoc or itemEquipLoc == "") then
					itemEquipLoc = self.locals.INVTYPE_OTHER
				end
				if localvars.items[i] == nil then
					localvars.items[i] = false
					localvars.count_known = localvars.count_known + 1
				end
				if (not types_tree[itemType]) then
					types_tree[itemType] = {}
					newtype = true
				end
				if (not types_tree[itemType][itemSubType]) then
					types_tree[itemType][itemSubType] = {}
					newtype = true
				end
				if (not types_tree[itemType][itemSubType][itemEquipLoc]) then
					types_tree[itemType][itemSubType][itemEquipLoc] = ""
					newtype = true
				end
			end
		end
	end
	if not newtype then return end
	for k in pairs(localvars.types_base) do
		localvars.types_base[k] = nil
	end
	for k in pairs(localvars.types_base) do
		localvars.types_base[k] = nil
	end
	for k in pairs(localvars.types_sub) do
		localvars.types_sub[k] = nil
	end
	for k in pairs(localvars.types_slot) do
		localvars.types_slot[k] = nil
	end
	for type, subtypes in pairs(localvars.types_tree) do
		table_insert(localvars.types_base, type)
		local subtypesort = {}
		local hasinvtypes = false
		for subtype, invtypes in pairs(subtypes) do
			table_insert(subtypesort, subtype)
			local invtypesort = {}
			for invtype, _ in pairs(invtypes) do
				table_insert(invtypesort, invtype)
			end
			if #invtypesort > 1 then
				table_sort(invtypesort)
				if (not localvars.types_slot[type]) then
					localvars.types_slot[type] = {}
				end
				localvars.types_slot[type][subtype] = invtypesort
				hasinvtypes = true
			end
		end
		if #subtypesort > 1 or hasinvtypes then
			table_sort(subtypesort)
			localvars.types_sub[type] = subtypesort
		end
	end
	table_sort(localvars.types_base)
end

function InventoryLib:StoreSessionItem(item, viacomm)

	local localvars = self.localvars

	item = getbaseid(item)
	if not item then return false end
	if (not localvars.items[item]) then
		if localvars.items[item] == nil then	-- item not mined from cache, yet
			local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc = GetItemInfo(item)
			if not itemName then	-- item not cached, yet
				Gratuity:SetHyperlink(FORMAT_itemid:format(item))
				localvars.items_rescan[item] = not viacomm and true or false
				return false
			end
			itemEquipLoc = _G[itemEquipLoc]
			if (not itemEquipLoc or itemEquipLoc == "") then
				itemEquipLoc = self.locals.INVTYPE_OTHER
			end
			localvars.items[item] = false
			localvars.count_known = localvars.count_known + 1
			local types_tree = localvars.types_tree
			if not (types_tree and types_tree[itemType] and types_tree[itemType][itemSubType] and types_tree[itemType][itemSubType][itemEquipLoc]) then -- at least one of the type strings isn't known yet so rebuild type trees
				self:BuildLists(false, item)
			end
		end
		localvars.items[item] = true
		if localvars.items_largestid < item then
			localvars.items_largestid = item
		end
		localvars.count_safe = localvars.count_safe + 1
	end
	return true
end

function InventoryLib:RecheckSessionItems()
	for item, viacomm in pairs(self.localvars.items_rescan) do
		if InventoryLib:StoreSessionItem(item, viacomm) then
			self.localvars.items_rescan[item] = nil
		end
	end
end

-- -----
-- the following functions scan various sources for items in order to add them to the list of linkable items for this session
-- -----

-- if given unit hasn't been inspected lately all slots will be scaned for items
function InventoryLib:Inspect(unit)
	if not (UnitExists(unit) and UnitIsPlayer(unit) and UnitIsFriend(unit, "player") and CheckInteractDistance(unit, 1)) then
		return
	end
	local timenow = time()
	if timenow - self.localvars.inspected_lastclean > 5 then -- clean up cache at most every 5 seconds
		for key, value in pairs(self.localvars.inspected_units) do
			if (timenow - value > 300) then -- remove after 5min
				self.localvars.inspected_units[key] = nil
			end
		end
		self.localvars.inspected_lastclean = timenow
	end
	local name, server = UnitName(unit)
	if not server then server = self.consts.RealmName end
	if self.localvars.inspected_units[name.."-"..server] then
		return
	end
	self.localvars.inspected_units[name.."-"..server] = timenow
	for _, slot in ipairs(self.consts.InventorySlotName) do
		local item = GetInventoryItemLink(unit, GetInventorySlotInfo(slot))
		if item then
			self:StoreSessionItem(item)
		end
	end
end

function InventoryLib:ScanAuctionPage()
	local numBatchAuctions = GetNumAuctionItems("list")
	for auctionid = 1, numBatchAuctions do
		local item = GetAuctionItemLink("list", auctionid)
		if item then
			self:StoreSessionItem(item)
		end
	end
end

function InventoryLib:ScanBags()
	for bagid = 0, NUM_BAG_SLOTS do
		local size = GetContainerNumSlots(bagid)
		if size then
			for slotid = 1, size do
				local item = GetContainerItemLink(bagid, slotid)
				if item then
					self:StoreSessionItem(item)
				end
			end
		end
	end
end

function InventoryLib:ScanBank()
	for bagid_i = NUM_BAG_SLOTS, (NUM_BAG_SLOTS + NUM_BANKBAGSLOTS) do
		local bagid = bagid_i
		if bagid == NUM_BAG_SLOTS then
			bagid = BANK_CONTAINER
		end
		local size = GetContainerNumSlots(bagid)
		if size then
			for slotid = 1, size do
				local item = GetContainerItemLink(bagid, slotid)
				if item then
					self:StoreSessionItem(item)
				end
			end
		end
	end
end

function InventoryLib:ScanMerchant()
	if MerchantFrame:IsVisible() and MerchantFrame.selectedTab == 1 then
		local numMerchantItems = GetMerchantNumItems()
		for i = 1, numMerchantItems do
			local item = GetMerchantItemLink(i)
			if item then
				self:StoreSessionItem(item)
			end
		end
	end
end

function InventoryLib:ScanTradeSkill()
	if TradeSkillFrame ~= nil and TradeSkillFrame:IsVisible() then
		for i=1, GetNumTradeSkills() do
			local _, type, _, _ = GetTradeSkillInfo(i)
			if type ~= "header" then
				local item = GetTradeSkillItemLink(i)
				if item then
					self:StoreSessionItem(item)
				end
				for j=1, GetTradeSkillNumReagents(i) do
					local reagentName, _, _, _ = GetTradeSkillReagentInfo(i, j)
					if reagentName then
						local item = GetTradeSkillReagentItemLink(i, j)
						if item then
							self:StoreSessionItem(item)
						end
					end
				end
			end
		end
	end
end

function InventoryLib:ScanCraft()
	if CraftFrame ~= nil and CraftFrame:IsVisible() then
		for i=1, GetNumCrafts() do
			local _, type, _, _ = GetCraftInfo(i)
			if type ~= "header" then
				local item = GetCraftItemLink(i)
				if item then
					self:StoreSessionItem(item)
				end
				for j=1, GetCraftNumReagents(i) do
					local reagentName, _, _, _ = GetCraftReagentInfo(i, j)
					if reagentName then
						local item = GetCraftReagentItemLink(i, j)
						if item then
							self:StoreSessionItem(item)
						end
					end
				end
			end
		end
	end
end

-- -----
-- lib stuff
-- -----

local function activate(self, oldLib, oldDeactivate)
	InventoryLib = self
	if oldLib then
		self.localvars = oldLib.localvars

		oldLib:CancelAllScheduledEvents()
		oldLib:UnregisterAllEvents()

		if oldLib.commPrefix then
			oldLib.UnregisterAllComms(self) -- AceLibrary's magic reference-changing
			self.commPrefix = oldLib.commPrefix
		end
	else
		self.localvars = {
			["framescan_last"]			= 0,			-- timestamp of latest frame scan (only one per second done)
			["inspected_lastclean"]		= 0,			-- timestamp of the last pruning of scanned units
			["inspected_units"]			= {},			-- list of latest scanned units
			["items"]					= {},			-- list of all known items keyed by id, true if session valid
			["items_scan_last"]			= 0,			-- id of the last scanned item; will scan consts.itemscancount items every second
			["items_largestid"]			= 0,			-- largest id of a valid item
			["items_send_first"]		= 1,			-- id of the next item to send
			["items_rescan"]			= {},			-- list of valid items not found in cache, yet
			["types_tree"]				= {},
			["types_base"]				= {},
			["types_sub"]				= {},
			["types_slot"]				= {},
			["count_known"]				= 0,
			["count_safe"]				= 0,
			["temp_table"]				= {},			-- table to be used inside functions; clear entries after use
			["known_users"]				= {},
			["known_users_time"]		= {},
			["known_users_rev"]			= {},
		}
	end

	self.consts = {
		-- list of all slots scanned on inspect
		["InventorySlotName"] = {
			"HeadSlot",
			"NeckSlot",
			"ShoulderSlot",
			"BackSlot",
			"ChestSlot",
			"ShirtSlot",
			"TabardSlot",
			"WristSlot",
			"HandsSlot",
			"WaistSlot",
			"LegsSlot",
			"FeetSlot",
			"Finger0Slot",
			"Finger1Slot",
			"Trinket0Slot",
			"Trinket1Slot",
			"MainHandSlot",
			"SecondaryHandSlot",
			"RangedSlot",
			"AmmoSlot",
			"Bag0Slot",
			"Bag1Slot",
			"Bag2Slot",
			"Bag3Slot",
		},
		["itemscancount"] = 100,
		["maxid"] = 40000,
		["comm_packetsize"] = 200,
		["RealmName"] = GetRealmName()
	}

	self.locals = {
		["INVTYPE_OTHER"] = "Other",
	}

	if ( GetLocale() == "deDE" ) then
		self.locals.INVTYPE_OTHER = "Sonstiges"
	end

	if oldDeactivate then
		oldDeactivate(oldLib)
	end
end

local function external(self, major, instance)
	if major == "AceEvent-2.0" then
		AceEvent = instance
		AceEvent:embed(self)
		self:UnregisterAllEvents()
		self:CancelAllScheduledEvents()
		if AceEvent:IsFullyInitialized() then
			self:AceEvent_FullyInitialized()
		else
			self:RegisterEvent("AceEvent_FullyInitialized", "AceEvent_FullyInitialized", true)
		end		
	elseif major == "AceComm-2.0" then
		instance:embed(self)
		if not self.commPrefix then
			self:SetCommPrefix("InvL")
		end
		self:RegisterComm(self.commPrefix, "GROUP")
		self:RegisterComm(self.commPrefix, "GUILD")
		self:RegisterComm(self.commPrefix, "WHISPER")
--[[	elseif major == "AceDebug-2.0" then
		instance:embed(self)
		self.debugFrame = ChatFrame4
		self:SetDebugging(true) ]]
	elseif major == "Gratuity-2.0" then
		Gratuity = instance
	end
end

-- -----
-- events
-- -----
local ChatMessageTypes = {
	"CHAT_MSG_SYSTEM",
	"CHAT_MSG_SAY",
	"CHAT_MSG_TEXT_EMOTE",
	"CHAT_MSG_YELL",
	"CHAT_MSG_WHISPER",
	"CHAT_MSG_WHISPER_INFORM",
	"CHAT_MSG_PARTY",
	"CHAT_MSG_GUILD",
	"CHAT_MSG_OFFICER",
	"CHAT_MSG_CHANNEL",
	"CHAT_MSG_RAID",
	"CHAT_MSG",
	"CHAT_MSG_LOOT",
	"CHAT_MSG_TRADE",
}

function InventoryLib:AceEvent_FullyInitialized()
	self:RegisterEvent("AUCTION_ITEM_LIST_UPDATE", "Event_Auction")
	self:RegisterEvent("BANKFRAME_OPENED", "Event_Bank")
	self:RegisterEvent("MERCHANT_SHOW", "Event_Merchant")
	self:RegisterBucketEvent("MERCHANT_UPDATE", 1, "Event_Merchant")
	self:RegisterEvent("TRADE_SKILL_SHOW", "Event_Tradeskill")
	self:RegisterBucketEvent("TRADE_SKILL_UPDATE", 1, "Event_Tradeskill")
	self:RegisterEvent("CRAFT_SHOW", "Event_Craft")
	self:RegisterBucketEvent("CRAFT_UPDATE", 1, "Event_Craft")

	self:ScheduleRepeatingEvent(self.BuildLists, 1, self)
	self:ScheduleRepeatingEvent(self.Schedule_2min, 120, self)
	self:ScheduleRepeatingEvent(self.ShareItems, 5, self)

	self:RegisterEvent("PLAYER_ENTERING_WORLD")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterEvent("PARTY_MEMBERS_CHANGED")
	self:RegisterEvent("UPDATE_MOUSEOVER_UNIT")

	for _, ChatMessageType in pairs(ChatMessageTypes) do
		self:RegisterEvent(ChatMessageType, "Event_Chat")
	end

	self:BuildLists(true)
	self:Inspect("player")
	self:ScanBags()

	self:AnnounceSelf("GUILD")
end

function InventoryLib:Event_Auction()
	self:ScanAuctionPage()
end

function InventoryLib:Event_Bank()
	self:Inspect("player")
	self:ScanBags()
	self:ScanBank()
end

function InventoryLib:Event_Merchant()
	self:ScanMerchant()
end

function InventoryLib:Event_Tradeskill()
	self:ScanTradeSkill()
end

function InventoryLib:Event_Craft()
	self:ScanCraft()
end

function InventoryLib:PLAYER_ENTERING_WORLD()
	self:Inspect("player")
	self:ScanBags()
end

function InventoryLib:PLAYER_TARGET_CHANGED()
	if UnitExists("target") and UnitIsPlayer("target") then
		self:Inspect("target")
	end
end

function InventoryLib:PARTY_MEMBERS_CHANGED()
	self:AnnounceSelf("GROUP")
end

function InventoryLib:UPDATE_MOUSEOVER_UNIT()
	if UnitExists("mouseover") and UnitIsPlayer("mouseover") then
		self:Inspect("mouseover")
	end
end

function InventoryLib:Event_Chat(msg)
	for id in msg:gmatch(PATTERN_itemlink) do
		self:StoreSessionItem(id)
	end
end

------------------------------------------------
-- AceComm
------------------------------------------------

local COMM_COMPAT = 2
InventoryLib.commHandlers = {}

function InventoryLib:OnCommReceive(prefix, sender, distribution, compat, handler, ...)
	if compat < COMM_COMPAT then
--~ 		self:SendCommMessage("WHISPER", sender, COMM_COMPAT, "OLDVERSION", MAJOR_VERSION, MINOR_VERSION)
		if type(handler) == 'string' and self.commHandlers[handler.."_"..compat] then
			self.commHandlers[handler.."_"..compat](self, sender, ...)
		end
	elseif compat > COMM_COMPAT then
		if handler == "OLDVERSION" then
			self.commHandlers[handler](self, sender, ...)
		else
			self.commHandlers:OLDVERSION(sender, "unknown", "")
		end
	elseif type(handler) == 'string' and self.commHandlers[handler] then
		self.commHandlers[handler](self, sender, ...)
	end
end

function InventoryLib.commHandlers:OLDVERSION(sender, major, minor)
	AceLibrary("AceConsole-2.0"):Print(("You have an older version of the InventoryLib library (%s - %s) than %s (%s - %s) -- please upgrade the addon which is using it."):format(MAJOR_VERSION, MINOR_VERSION, sender, major, minor))
end

-- handler for data from COMM_COMPAT 1
function InventoryLib.commHandlers:VI_1(sender, ...) -- validate items
	for i = 1, select('#', ...) do
		self:StoreSessionItem(select(i, ...), true)
	end
end

function InventoryLib.commHandlers:HELLO(sender, revision)
	local known_users = self.localvars.known_users
	local known_users_time = self.localvars.known_users_time
	local known_users_rev = self.localvars.known_users_rev
	if strfind(sender, "-") then return end
	if not known_users_time[sender] then
		table_insert(known_users, sender)
		table_sort(known_users)
	end
	known_users_time[sender] = time()
	known_users_rev[sender] = revision
end

local shifttable = {1,2,4,8,16,32,64,128}
function InventoryLib.commHandlers:VI(sender, itemstring) -- validate items
-- *****
	local items = self.localvars.items
-- *****
	if strlen(itemstring) < 4 then return end
	-- first three byte are the id of the first item in packet
	local base = strbyte(itemstring, 1) * 65536 + strbyte(itemstring, 2) * 256 + strbyte(itemstring, 3) - 33 -- first itemid - (4 byte + 1 bit offset)

	for ibyte = 4, strlen(itemstring) do
		local dbyte = strbyte(itemstring, ibyte)
		for ibit = 1, 8 do
			if (band(dbyte, shifttable[ibit]) ~= 0) then -- the bit is set -> the item is valid
				local itemid = base + ibyte * 8 + ibit
				self:StoreSessionItem(itemid, true)
-- *****
--~ 				if not items[itemid] then
--~ 					AceLibrary("AceConsole-2.0"):Print(("recived %d from %s, which is not known safe, yet."):format(itemid, sender))
--~ 				end
-- *****
			end
		end
	end
end

function InventoryLib:Schedule_2min()
	self:CleanKnownUsers()
	self:AnnounceSelf()
end

function InventoryLib:CleanKnownUsers()
	local known_users = self.localvars.known_users
	local known_users_time = self.localvars.known_users_time
	local timenow = time()

	for i = #known_users, 1, -1 do
		if known_users_time[known_users[i]] + 600 < timenow then
			known_users_time[known_users[i]] = nil
			table_remove(known_users, i)
		end
	end
end

function InventoryLib:AnnounceSelf(channel)
	if not self.SendCommMessage then return end
	local revision = MINOR_VERSION:match("(%d+)")
	if channel == "GROUP" then
		self:SendCommMessage("GROUP", COMM_COMPAT, "HELLO", revision)
	elseif channel == "GUILD" then
		self:SendCommMessage("GUILD", COMM_COMPAT, "HELLO", revision)
	else
		for _, target in ipairs(self.localvars.known_users) do
			self:SendCommMessage("WHISPER", target, COMM_COMPAT, "HELLO", revision)
		end
	end
end

local outtable = {}
function InventoryLib:ShareItems()
	if not self.SendCommMessage then return end
	local known_users_list = self.localvars.known_users_list
	local items = self.localvars.items
	local COMM_PACKETSIZE = self.consts.comm_packetsize
--~ 	if #known_users_list <= 0 then return end
	
	local items_send_first = self.localvars.items_send_first
	if items_send_first > self.localvars.items_largestid then items_send_first = 1 end
	if #outtable > COMM_PACKETSIZE + 3 then
		for i = #outtable, COMM_PACKETSIZE + 4, -1 do
			outtable[i] = nil
		end
	end
	outtable[1] = strchar(math_floor(items_send_first / 256^2) % 256)
	outtable[2] = strchar(math_floor(items_send_first / 256)   % 256)
	outtable[3] = strchar(           items_send_first          % 256)
	local hasdata = false
	for i = 0, COMM_PACKETSIZE - 1 do
		local base = items_send_first + 8 * i
		local outchar = (items[base] and 1 or 0) + (items[base + 1] and 2 or 0) + (items[base + 2] and 4 or 0) + (items[base + 3] and 8 or 0) + (items[base + 4] and 16 or 0) + (items[base + 5] and 32 or 0) + (items[base + 6] and 64 or 0) + (items[base + 7] and 128 or 0)
		hasdata = hasdata or (outchar ~= 0)
		outtable[i+4] = strchar(outchar)
	end
	self.localvars.items_send_first = items_send_first + COMM_PACKETSIZE * 8
	if not hasdata then return end

	local itemdata = table_concat(outtable)
	local target = known_users[math_random(#known_users_list)]
-- *****
--~ 	local target = UnitName("player")
-- *****
	self:SendCommMessage("WHISPER", target, COMM_COMPAT, "VI", itemdata)
end

-- -----
-- load the lib
-- -----
AceLibrary:Register(InventoryLib, MAJOR_VERSION, MINOR_VERSION, activate, nil, external)
